"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const util_1 = require("util");
const constants_1 = require("../utils/constants");
const closeAsync = util_1.promisify(fs_1.close);
class FDCache {
    constructor(size = constants_1.DEFAULT_FD_CACHE_NUMBER) {
        this.initialized = false;
        this.closed = false;
        if (size < constants_1.FD_CACHE_NUMBER_MIN || size > constants_1.FD_CACHE_NUMBER_MAX) {
            size = constants_1.DEFAULT_FD_CACHE_NUMBER;
        }
        this.size = size;
        this.queue = [];
        this.cache = new Map();
    }
    /**
     * Init the FDCache.
     *
     * @returns {Promise<void>}
     * @memberof IFDCache
     */
    async init() {
        this.initialized = true;
    }
    /**
     * Close FDCache, release the file descriptor.
     *
     * @returns {Promise<void>}
     * @memberof IFDCache
     */
    async close() {
        await this.clear();
        this.closed = true;
    }
    /**
     * Whether the fd has been initialized.
     *
     * @returns {boolean}
     * @memberof IFDCache
     */
    isInitialized() {
        return this.initialized;
    }
    /**
     * Whether the fd has benn closed.
     *
     * @returns {boolean}
     * @memberof IFDCache
     */
    isClosed() {
        return this.closed;
    }
    /**
     * Get the fd with the given id.
     * Return -1 if the fd does not exist in the cache.
     *
     * @param {string} id
     * @returns {(number | undefined)}
     * @memberof FDCache
     */
    get(id) {
        const fd = this.cache.get(id);
        return fd;
    }
    /**
     * Insert a new fd to the cache.
     * Replace an old item if the cache is full. The replaced FD should be closed.
     *
     * @param {string} id
     * @param {number} fd
     * @returns {void}
     * @memberof IFDCache
     */
    insert(id, fd) {
        const count = this.queue.length;
        if (count === this.size) {
            const head = this.queue.shift();
            const cachedfd = this.cache.get(head);
            if (cachedfd !== undefined) {
                closeAsync(cachedfd);
                this.cache.delete(head);
            }
        }
        this.queue.push(id);
        this.cache.set(id, fd);
    }
    async clear() {
        const closeFDPromises = [];
        this.cache.forEach(fd => {
            closeFDPromises.push(closeAsync(fd));
        });
        await Promise.all(closeFDPromises);
        this.queue = [];
        this.cache.clear();
    }
}
exports.default = FDCache;
//# sourceMappingURL=FDCache.js.map