"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_1 = require("fs");
const multistream = require("multistream");
const path_1 = require("path");
const util_1 = require("util");
const uuid = require("uuid");
const IBlobMetadataStore_1 = require("../../blob/persistence/IBlobMetadataStore");
const BufferStream_1 = tslib_1.__importDefault(require("../utils/BufferStream"));
const constants_1 = require("../utils/constants");
const utils_1 = require("../utils/utils");
const ZeroBytesStream_1 = tslib_1.__importDefault(require("../ZeroBytesStream"));
const FDCache_1 = tslib_1.__importDefault(require("./FDCache"));
const OperationQueue_1 = tslib_1.__importDefault(require("./OperationQueue"));
const statAsync = util_1.promisify(fs_1.stat);
const mkdirAsync = util_1.promisify(fs_1.mkdir);
const unlinkAsync = util_1.promisify(fs_1.unlink);
// The max size of an extent.
const MAX_EXTENT_SIZE = constants_1.DEFAULT_MAX_EXTENT_SIZE;
var AppendStatusCode;
(function (AppendStatusCode) {
    AppendStatusCode[AppendStatusCode["Idle"] = 0] = "Idle";
    AppendStatusCode[AppendStatusCode["Appending"] = 1] = "Appending";
})(AppendStatusCode || (AppendStatusCode = {}));
// const openAsync = promisify(open);
/**
 * Persistency layer data store source implementation interacting with the storage media.
 * It provides the methods to read and write data with the storage.
 *
 * @export
 * @class FSExtentStore
 * @implements {IExtentStore}
 */
class FSExtentStore {
    constructor(metadata, persistencyConfiguration, logger) {
        this.persistencyConfiguration = persistencyConfiguration;
        this.logger = logger;
        this.initialized = false;
        this.closed = false;
        this.appendExtentArray = [];
        this.persistencyPath = new Map();
        this.fdCache = new FDCache_1.default(100);
        for (const storeDestination of persistencyConfiguration) {
            this.persistencyPath.set(storeDestination.persistencyId, storeDestination.persistencyPath);
            for (let i = 0; i < storeDestination.maxConcurrency; i++) {
                const appendExtent = this.createAppendExtent(storeDestination.persistencyId);
                this.appendExtentArray.push(appendExtent);
            }
        }
        this.appendExtentNumber = this.appendExtentArray.length;
        this.metadataStore = metadata;
        this.appendQueue = new OperationQueue_1.default(this.appendExtentNumber, logger);
        // TODO:Should add to interface to pass this parameter.
        this.readQueue = new OperationQueue_1.default(constants_1.DEFAULT_READ_CONCURRENCY, logger);
    }
    isInitialized() {
        return this.initialized;
    }
    isClosed() {
        return this.closed;
    }
    async init() {
        for (const storeDestination of this.persistencyConfiguration) {
            try {
                await statAsync(storeDestination.persistencyPath);
            }
            catch (_a) {
                await mkdirAsync(storeDestination.persistencyPath);
            }
        }
        if (!this.metadataStore.isInitialized()) {
            await this.metadataStore.init();
        }
        if (!this.fdCache.isInitialized()) {
            await this.fdCache.init();
        }
        this.initialized = true;
        this.closed = false;
    }
    async close() {
        if (!this.fdCache.isClosed()) {
            await this.fdCache.close();
        }
        if (!this.metadataStore.isClosed()) {
            await this.metadataStore.close();
        }
        this.closed = true;
    }
    async clean() {
        if (this.isClosed()) {
            for (const path of this.persistencyConfiguration) {
                try {
                    await utils_1.rimrafAsync(path.persistencyPath);
                }
                catch (_a) {
                    // TODO: Find out why sometimes it throws no permission error
                    /* NOOP */
                }
            }
            return;
        }
        throw new Error(`Cannot clean FSExtentStore, it's not closed.`);
    }
    /**
     * This method may create a new extent or append data to an existing extent.
     * Return the extent chunk information including the extentId, offset and count.
     *
     * @param {(NodeJS.ReadableStream | Buffer)} data
     * @param {string} [contextId]
     * @returns {Promise<IExtentChunk>}
     * @memberof FSExtentStore
     */
    async appendExtent(data, contextId) {
        const op = () => new Promise((resolve, reject) => {
            (async () => {
                let appendExtentIdx = 0;
                for (let i = 1; i < this.appendExtentNumber; i++) {
                    if (this.appendExtentArray[i].appendStatus === AppendStatusCode.Idle) {
                        appendExtentIdx = i;
                        break;
                    }
                }
                this.appendExtentArray[appendExtentIdx].appendStatus =
                    AppendStatusCode.Appending;
                this.logger.info(`FSExtentStore:appendExtent() Select extent from idle location for extent append operation. LocationId:${appendExtentIdx} extentId:${this.appendExtentArray[appendExtentIdx].id} offset:${this.appendExtentArray[appendExtentIdx].offset} MAX_EXTENT_SIZE:${MAX_EXTENT_SIZE} `, contextId);
                if (this.appendExtentArray[appendExtentIdx].offset > MAX_EXTENT_SIZE) {
                    this.getNewExtent(this.appendExtentArray[appendExtentIdx]);
                    this.logger.info(`FSExtentStore:appendExtent() Size of selected extent offset is larger than maximum extent size ${MAX_EXTENT_SIZE} bytes, try appending to new extent. New extent LocationID:${appendExtentIdx} extentId:${this.appendExtentArray[appendExtentIdx].id} offset:${this.appendExtentArray[appendExtentIdx].offset} MAX_EXTENT_SIZE:${MAX_EXTENT_SIZE} `, contextId);
                }
                let rs;
                if (data instanceof Buffer) {
                    rs = new BufferStream_1.default(data);
                }
                else {
                    rs = data;
                }
                const appendExtent = this.appendExtentArray[appendExtentIdx];
                const id = appendExtent.id;
                const path = this.generateExtentPath(appendExtent.persistencyId, id);
                let fd = this.fdCache.get(id);
                this.logger.debug(`FSExtentStore:appendExtent() Get fd:${fd} for extent:${id} from cache.`, contextId);
                if (fd === undefined) {
                    fd = fs_1.openSync(path, "a"); // TODO: async
                    this.fdCache.insert(id, fd);
                    this.logger.debug(`FSExtentStore:appendExtent() Open file:${path} for extent:${id}, get new fd:${fd}`, contextId);
                }
                const ws = fs_1.createWriteStream(path, {
                    flags: "a",
                    fd,
                    autoClose: false
                });
                let count = 0;
                this.logger.debug(`FSExtentStore:appendExtent() Start writing to extent ${id}`, contextId);
                try {
                    count = await this.streamPipe(rs, ws);
                    const offset = appendExtent.offset;
                    appendExtent.offset += count;
                    const extent = {
                        id,
                        persistencyId: appendExtent.persistencyId,
                        path: id,
                        size: count + offset,
                        lastModifiedInMS: Date.now()
                    };
                    this.logger.debug(`FSExtentStore:appendExtent() Write finish, start updating extent metadata. extent:${JSON.stringify(extent)}`, contextId);
                    await this.metadataStore.updateExtent(extent);
                    this.logger.debug(`FSExtentStore:appendExtent() Update extent metadata done. Resolve()`, contextId);
                    appendExtent.appendStatus = AppendStatusCode.Idle;
                    return {
                        id,
                        offset,
                        count
                    };
                }
                catch (err) {
                    appendExtent.appendStatus = AppendStatusCode.Idle;
                    throw err;
                }
            })()
                .then(resolve)
                .catch(reject);
        });
        return this.appendQueue.operate(op, contextId);
    }
    /**
     * Read data from persistency layer according to the given IExtentChunk.
     *
     * @param {IExtentChunk} [extentChunk]
     * @returns {Promise<NodeJS.ReadableStream>}
     * @memberof FSExtentStore
     */
    async readExtent(extentChunk, contextId) {
        if (extentChunk === undefined || extentChunk.count === 0) {
            return new ZeroBytesStream_1.default(0);
        }
        if (extentChunk.id === IBlobMetadataStore_1.ZERO_EXTENT_ID) {
            const subRangeCount = Math.min(extentChunk.count);
            return new ZeroBytesStream_1.default(subRangeCount);
        }
        const persistencyId = await this.metadataStore.getExtentPersistencyId(extentChunk.id);
        const path = this.generateExtentPath(persistencyId, extentChunk.id);
        const op = () => new Promise((resolve, reject) => {
            this.logger.verbose(`FSExtentStore:readExtent() Creating read stream. LocationId:${persistencyId} extentId:${extentChunk.id} path:${path} offset:${extentChunk.offset} count:${extentChunk.count} end:${extentChunk.offset + extentChunk.count - 1}`, contextId);
            const stream = fs_1.createReadStream(path, {
                start: extentChunk.offset,
                end: extentChunk.offset + extentChunk.count - 1
            }).on("close", () => {
                this.logger.verbose(`FSExtentStore:readExtent() Read stream closed. LocationId:${persistencyId} extentId:${extentChunk.id} path:${path} offset:${extentChunk.offset} count:${extentChunk.count} end:${extentChunk.offset + extentChunk.count - 1}`, contextId);
            });
            resolve(stream);
        });
        return this.readQueue.operate(op, contextId);
    }
    /**
     * Merge several extent chunks to a ReadableStream according to the offset and count.
     *
     * @param {(IExtentChunk)[]} extentChunkArray
     * @param {number} [offset=0]
     * @param {number} [count=Infinity]
     * @param {string} [contextId]
     * @returns {Promise<NodeJS.ReadableStream>}
     * @memberof FSExtentStore
     */
    async readExtents(extentChunkArray, offset = 0, count = Infinity, contextId) {
        this.logger.verbose(`FSExtentStore:readExtents() Start read from multi extents...`, contextId);
        if (count === 0) {
            return new ZeroBytesStream_1.default(0);
        }
        const start = offset; // Start inclusive position in the merged stream
        const end = offset + count; // End exclusive position in the merged stream
        const streams = [];
        let accumulatedOffset = 0; // Current payload offset in the merged stream
        for (const chunk of extentChunkArray) {
            const nextOffset = accumulatedOffset + chunk.count;
            if (nextOffset <= start) {
                accumulatedOffset = nextOffset;
                continue;
            }
            else if (end <= accumulatedOffset) {
                break;
            }
            else {
                let chunkStart = chunk.offset;
                let chunkEnd = chunk.offset + chunk.count;
                if (start > accumulatedOffset) {
                    chunkStart = chunkStart + start - accumulatedOffset; // Inclusive
                }
                if (end <= nextOffset) {
                    chunkEnd = chunkEnd - (nextOffset - end); // Exclusive
                }
                streams.push(await this.readExtent({
                    id: chunk.id,
                    offset: chunkStart,
                    count: chunkEnd - chunkStart
                }, contextId));
                accumulatedOffset = nextOffset;
            }
        }
        // TODO: What happens when count exceeds merged payload length?
        // throw an error of just return as much data as we can?
        if (end !== Infinity && accumulatedOffset < end) {
            throw new RangeError(
            // tslint:disable-next-line:max-line-length
            `Not enough payload data error. Total length of payloads is ${accumulatedOffset}, while required data offset is ${offset}, count is ${count}.`);
        }
        return multistream(streams);
    }
    /**
     * Delete the extents from persistency layer.
     *
     * @param {Iterable<string>} persistency
     * @returns {Promise<void>}
     * @memberof IExtentStore
     */
    async deleteExtents(extents) {
        for (const id of extents) {
            const persistencyId = await this.metadataStore.getExtentPersistencyId(id);
            const path = this.generateExtentPath(persistencyId, id);
            try {
                await unlinkAsync(path);
                await this.metadataStore.deleteExtent(id);
            }
            catch (err) {
                if (err.code === "ENOENT") {
                    await this.metadataStore.deleteExtent(id);
                }
            }
        }
        return;
    }
    /**
     * Return its metadata store.
     *
     * @returns {IExtentMetadataStore}
     * @memberof IExtentStore
     */
    getMetadataStore() {
        return this.metadataStore;
    }
    async streamPipe(rs, ws) {
        return new Promise((resolve, reject) => {
            let count = 0;
            let wsEnd = false;
            rs.on("data", data => {
                count += data.length;
                if (!ws.write(data)) {
                    rs.pause();
                }
            })
                .on("end", () => {
                if (!wsEnd) {
                    ws.end();
                    wsEnd = true;
                }
            })
                .on("close", () => {
                if (!wsEnd) {
                    ws.end();
                    wsEnd = true;
                }
            })
                .on("error", err => {
                ws.destroy(err);
            });
            ws.on("drain", () => {
                rs.resume();
            })
                .on("finish", () => {
                resolve(count);
            })
                .on("error", reject);
        });
    }
    /**
     * Create a new append extent model for a new write directory.
     *
     * @private
     * @param {string} persistencyPath
     * @returns {IAppendExtent}
     * @memberof FSExtentStore
     */
    createAppendExtent(persistencyId) {
        return {
            id: uuid(),
            offset: 0,
            appendStatus: AppendStatusCode.Idle,
            persistencyId
        };
    }
    /**
     * Select a new extent to append for an exist write directory.
     *
     * @private
     * @param {IAppendExtent} appendExtent
     * @memberof FSExtentStore
     */
    getNewExtent(appendExtent) {
        appendExtent.id = uuid();
        appendExtent.offset = 0;
    }
    /**
     * Generate the file path for a new extent.
     *
     * @private
     * @param {string} extentId
     * @returns {string}
     * @memberof FSExtentStore
     */
    generateExtentPath(persistencyId, extentId) {
        const directoryPath = this.persistencyPath.get(persistencyId);
        if (!directoryPath) {
            // To be completed
        }
        return path_1.join(directoryPath, extentId);
    }
}
exports.default = FSExtentStore;
//# sourceMappingURL=FSExtentStore.js.map